////////////////////////////////////////////////////////////////////////////////
//
// Fracter - A simple Mandelbrot Set viewer.
//
// Copyright (C) 2010 Sarah Eris Horsley Caffee
//
// This file is part of Fracter.
//
// Fracter is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//
// App.cpp
// Application class implementation
//
////////////////////////////////////////////////////////////////////////////////

#include <iomanip>
#include <iostream>
#include <sstream>
#include <typeinfo>
#include <vector>

#if defined(WIN32)
#include <cstdlib>
#include <windows.h>
#include <Shlwapi.h>

#else
#include <cctype>
#include <cstring>
#include <cstdio>

#include <dirent.h>
#endif


#include "SDL.h"

#include "App.h"

#include "Options.h"
#include "Mandelbrot.h"
#include "Julia.h"


////////////////////////////////////////////////////////////////////////////////

App::App() :
   running (true),
   redraw (false),
   window (NULL),
   surf_selection (NULL),
   surf_fractal (NULL),
   selection_x (-1),
   selection_y (-1),
   selection_width (-1),
   setting_zoom (false),
   can_set_julia_k (false),
   setting_julia_k (false),
   old_views (),
   fractal (NULL)
   {
   }


////////////////////////////////////////////////////////////////////////////////
App::~App()
   {
   old_views.clear();
   }


////////////////////////////////////////////////////////////////////////////////

int App::run()
   {
   if (init() == false)
      return 1;

   SDL_Event event;

   while (running)
      {
      SDL_WaitEvent(&event);
      do
	 {
	 on_Event(&event);
	 }
      while (SDL_PollEvent(&event));

      update();

      if (redraw)
	 render();
      }

   cleanup();

   return 0;
   }


////////////////////////////////////////////////////////////////////////////////

bool App::init()
   {
   if (!init_sdl())
      {
      return false;
      }

   if (!create_new_surface(surf_selection)) { return false; }
   if (!create_new_surface(surf_fractal))   { return false; }


   std::cerr << "Available fractal types are" << std::endl;
   std::map <std::string, FractalFactory *> mymap = get_Fractal_map();
   std::map <std::string, FractalFactory *>::iterator it;
   for ( it=mymap.begin() ; it != mymap.end(); it++ )
      std::cerr << "  " << (*it).first << std::endl;

   FractalFactory * fractal_factory = 
	 get_Fractal_map()[typeid(Mandelbrot).name()];
   if (!fractal_factory)
      {
      std::cerr << "Mandelbrot set not available!" << std::endl;
      return false;
      }
   fractal = fractal_factory->create();
   fractal->init(surf_fractal);
   redraw = true;

   can_set_julia_k = true;

   set_caption();

   return true;
   }


////////////////////////////////////////////////////////////////////////////////

void App::update()
   {
   }


////////////////////////////////////////////////////////////////////////////////

void App::render()
   {
   if (redraw)
      {
      // Start with a clean slate.
      SDL_SetRenderDrawColor(0, 0, 0, SDL_ALPHA_OPAQUE);
      SDL_RenderClear();

      fractal->calc_set();
      SDL_BlitSurface(surf_fractal, NULL, surf_display, NULL);

      // draw a selection box
      if (setting_zoom)
	 {
	 SDL_BlitSurface(surf_selection, NULL, surf_display, NULL);
	 }

      // Show it to the user.
      SDL_RenderPresent();

      redraw = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::cleanup()
   {

   if (surf_selection)
      {
      SDL_FreeSurface(surf_selection);
      surf_selection = NULL;
      }
   if (window)
      {
      SDL_DestroyWindow(window);
      window = NULL;
      }

   SDL_Quit();
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_InputFocus()
   {
   SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);
   SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_ENABLE);
   SDL_EventState(SDL_MOUSEBUTTONUP, SDL_ENABLE);
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_InputBlur()
   {
   SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
   SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
   SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_KeyDown(SDL_KeyboardEvent key)
   {
   switch (key.keysym.sym)
      {
      case SDLK_ESCAPE:
      case SDLK_q:
	 SDL_Event event;
	 event.type = SDL_QUIT;
	 SDL_PushEvent(&event);
	 break;

      case SDLK_b:
      	 restore_view();
	 redraw = true;
	 set_caption();
      	 break;

      case SDLK_s:
      	 save_image();
      	 break;

      case SDLK_UP:
	 fractal->inc_max_iter();
	 set_caption();
	 break;
      case SDLK_DOWN:
	 fractal->dec_max_iter();
	 set_caption();
	 break;


      case SDLK_m:
	 if (fractal)
	    {
	    FractalFactory * fractal_factory = 
		  get_Fractal_map()[typeid(Mandelbrot).name()];
	    if (!fractal_factory)
	       {
	       std::cerr << "Mandelbrot set not available!" << std::endl;
	       SDL_Event event;
	       event.type = SDL_QUIT;
	       SDL_PushEvent(&event);
	       }
	    else
	       {
	       save_view();
	       delete fractal;
	       fractal = fractal_factory->create();
	       fractal->init(surf_fractal);
	       redraw = true;
	       set_caption();
	       can_set_julia_k = true;
	       }
	    }
	 break;

      case SDLK_j:
	 if (fractal)
	    {
	    FractalFactory * fractal_factory = 
		  get_Fractal_map()[typeid(Julia).name()];
	    if (!fractal_factory)
	       {
	       std::cerr << "Julia set not available!" << std::endl;
	       }
	    else
	       {
	       save_view();
	       delete fractal;
	       fractal = fractal_factory->create();
	       fractal->init(surf_fractal);
	       redraw = true;
	       set_caption();
	       can_set_julia_k = false;
	       }
	    }
	 break;

      case SDLK_k:
	 if (can_set_julia_k)
	    {
	    if (setting_julia_k) setting_julia_k = false;
	    else setting_julia_k = true;
	    }
	 break;

      default:
	 break;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_KeyUp(SDL_KeyboardEvent key)
   {
   switch (key.keysym.sym)
      {
      case SDLK_UP:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      case SDLK_DOWN:
	 fractal->set_calc_needed();
	 redraw = true;
	 break;
      default:
	 break;
      }   
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_MouseMove(SDL_MouseMotionEvent motion)
   {
   if (setting_zoom)
      {
      Uint32 clear = SDL_MapRGBA(surf_selection->format, 
				 0, 0, 0, SDL_ALPHA_TRANSPARENT);
      Uint32 white = SDL_MapRGBA(surf_selection->format, 
				 255, 255, 255, SDL_ALPHA_OPAQUE);

      SDL_FillRect(surf_selection, &surf_selection->clip_rect, clear);

      SDL_Rect rect;
      rect.x = selection_x;
      rect.y = selection_y;
      int w = mouse_move_width(motion.x, motion.y);
      rect.w = rect.h = w;

      SDL_FillRect(surf_selection, &rect, white);

      rect.x = selection_x + 1;
      rect.y = selection_y + 1;
      rect.w = rect.h = w - 2;

      SDL_FillRect(surf_selection, &rect, clear);

      redraw = true;
      }
   }

////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonDown(SDL_MouseButtonEvent button)
   {
   if (setting_julia_k)
      {
      set_julia_k(button.x, button.y);
      }
   else
      {
      selection_x = button.x;
      selection_y = button.y;
      setting_zoom = true;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_LButtonUp(SDL_MouseButtonEvent button)
   {
   if (setting_zoom)
      {
      selection_width = mouse_move_width(button.x, button.y);
      if ( (button.x <= 0) || 
	   (button.x >= surf_display->w - 1) || 
	   (button.y <= 0) || 
	   (button.y >= surf_display->h - 1) ||
	   (selection_width == 0))
	 {
	 // Selection cancelled
	 setting_zoom = false;
	 return;
	 }

      // calculate new min/max re/im
      double Re_scale = (fractal->get_size())/(surf_fractal->w - 1);
      double Im_scale = (fractal->get_size())/(surf_fractal->h - 1);

      double new_re_min = fractal->get_re_min() + (selection_x) * Re_scale;
      double new_im_max = fractal->get_im_max() - (selection_y) * Im_scale;
      double new_size = selection_width * Re_scale;

      save_view();

      fractal->set_re_min(new_re_min);
      fractal->set_im_max(new_im_max);
      fractal->set_size(new_size);
      fractal->set_calc_needed();
      redraw = true;

      set_caption();

      setting_zoom = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::on_Expose()
   {
   redraw = true;
   }


////////////////////////////////////////////////////////////////////////////////

void App::on_Quit()
   {
   running = false;
   }


////////////////////////////////////////////////////////////////////////////////
bool App::create_new_surface(SDL_Surface * &surface)
   {
   if (surface)
      {
      SDL_FreeSurface(surface);
      surface = NULL;
      }

   SDL_Surface * tmp;
   if ((tmp = SDL_CreateRGBSurface(SDL_HWSURFACE|SDL_SRCALPHA, 
				   Options::width, Options::height, 
				   Options::bpp, 0, 0, 0, 0))
       == NULL)
      {
      std::cerr << "SDL_CreateRGBSurface() failed in "
		<< "App::create_new_surface(): " 
		<< SDL_GetError() << std::endl;
      return false;
      }

   surface = SDL_DisplayFormatAlpha(tmp);
   SDL_FreeSurface(tmp);

   return true;
   }


////////////////////////////////////////////////////////////////////////////////
bool init_sdl()
   {
   if ((SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER)) == -1)
      {
      std::cerr << "SDL_Init() failed: " << SDL_GetError() << std::endl;
      return false;
      }

   if ((window = SDL_CreateWindow("SDL 1.3 test", 
				  SDL_WINDOWPOS_UNDEFINED, 
				  SDL_WINDOWPOS_UNDEFINED,
				  Options::width, Options::height,
				  0)) == 0)
      {
      std::cerr << "SDL_CreateWindow() failed: " 
   		<< SDL_GetError() << std::endl;
      return false;
      }

   // Go through the available renderers and find the one we need.
   // Do not use the x11 renderer.  At least for now it is not working
   // correctly..
   int max_drivers = SDL_GetNumRenderDrivers();
   if (max_drivers < 1)
      {
      std::cerr << "SDL_GetNumRenderDrivers() failed: "
		<< SDL_GetError() << std::endl;
      return false;
      }

   int driver = -1;
   for(int index = 0; index < max_drivers ; index++) 
      { 
      SDL_RendererInfo info; 
      if(SDL_GetRenderDriverInfo(index, &info) != -1)
	 { 
	 if (strncmp(info.name, "software", 8) == 0) driver = index;
	 std::cerr << "Driver " << info.name << " index: "
		   << index << std::endl; 
	 } 
      else
	 {
	 std::cerr << "SDL_GetRenderDriverInfo() for index " << index 
		   << " failed: " 
		   << SDL_GetError() << std::endl;
	 return false;
	 }
      } 

   if (driver == -1)
      {
      std::cerr << "SDL software renderer not available!" << std::endl;
      return false;
      }

   if (SDL_CreateRenderer(window, -1, 
			  SDL_RENDERER_PRESENTFLIP2 | 
			  SDL_RENDERER_ACCELERATED) != 0)
      {
      std::cerr << "SDL_CreateRenderer() failed: " 
   		<< SDL_GetError() << std::endl;
      return false;
      }

   SDL_RendererInfo info;
   if (SDL_GetRendererInfo(&info) == -1)
      {
      std::cerr << "SDL_GetRendererInfo() failed: "
		<< SDL_GetError() << std::endl;
      return false;
      }
   std::cerr << "Renderer: " << info.name << std::endl;


   if (SDL_AddTimer(10, App::timer_callback, NULL) == NULL)
      {
      std::cerr << "SDL_AddTimer() failed: "
		<< SDL_GetError() << std::endl;
      return false;
      }

   return true;
   }


////////////////////////////////////////////////////////////////////////////////
int App::mouse_move_width(int mx, int my)
   {
   int move_width = (mx - selection_x);
   if (mx > selection_x)
      {
      move_width *= -1;
      }
   
   int tmp_i = (my - selection_y);
   if (my < selection_y)
      {
      tmp_i *= -1;
      }
   
   if (tmp_i > move_width)
      move_width = tmp_i;

   return move_width;
   }


////////////////////////////////////////////////////////////////////////////////
void App::set_caption()
   {
   std::stringstream ss;
   ss << fractal->get_display_name() << " / "
      << "Real: " << fractal->get_re_min()
       << " to "<< fractal->get_re_min() + fractal->get_size()
      << " / Imaginary: " << fractal->get_im_max() - fractal->get_size()
      << " to " << fractal->get_im_max()
      << " / Max iterations: " << fractal->get_max_iter();
   SDL_SetWindowTitel(window, ss.str().c_str());
   }


////////////////////////////////////////////////////////////////////////////////
Uint32 App::timer_callback(Uint32 interval, void *param)
   {
   SDL_Event e;
 
   e.type = SDL_USEREVENT;
   e.user.code = 0;
   e.user.data1 = NULL;
   e.user.data2 = NULL;
   SDL_PushEvent(& e);
 
   return interval;
   }


////////////////////////////////////////////////////////////////////////////////
void App::set_julia_k(int x, int y)
   {
   FractalFactory * fractal_factory = get_Fractal_map()[typeid(Julia).name()];
   if (!fractal_factory)
      {
      std::cerr << "Julia set not available!" << std::endl;
      return;
      }
   else
      {
      save_view();
   
      delete fractal;
      fractal = fractal_factory->create();
      fractal->init(surf_fractal);

      double k[2];
      k[0]= fractal->get_re_min() + x * 
	    (fractal->get_size())/(surf_fractal->w - 1);
      k[1] = fractal->get_im_max() - y * 
	    (fractal->get_size())/(surf_fractal->h - 1);
      
      fractal->set_option("k", k);
      fractal->set_calc_needed();
      redraw = true;
      
      set_caption();
      setting_julia_k = false;
      }
   }


////////////////////////////////////////////////////////////////////////////////

void App::restore_view()
   {
   if (old_views.size() > 0)
      {
      View * v = old_views.back();
      old_views.pop_back();

      if (v->get_type().compare(fractal->get_fractal_name()) != 0)
	 {
	 delete fractal;
	 FractalFactory * fractal_factory = get_Fractal_map()[v->get_type()];
	 if (!fractal_factory)
	    {
	    std::cerr << "Cannot restore old view: unknown Fractal type " 
		      << v->get_type() << std::endl;
	    SDL_Event event;
	    event.type = SDL_QUIT;
	    SDL_PushEvent(&event);
	    return;
	    }
	 fractal = fractal_factory->create();
	 fractal->init(surf_fractal);
	 }

      fractal->restore_view(v);
      delete v;
      }
   }


////////////////////////////////////////////////////////////////////////////////
void App::save_view()
   {
   View * v = fractal->save_view();
   old_views.push_back(v);
   }


////////////////////////////////////////////////////////////////////////////////
bool App::save_image()
   {
   int i = App::get_next_file_num();
   std::stringstream name;
   name << "fracter-" << std::setw(3) << std::setfill('0') << i << ".bmp";

   if (SDL_SaveBMP(surf_display, name.str().c_str()) != 0)
      {
      std::cerr << "Unable to save image file " << name.str() << "" << std::endl;
      return false;
      }

   std::cerr << "Saved image as " << name.str() << std::endl;

   return true;
   }


#if defined(WIN32)
////////////////////////////////////////////////////////////////////////////////
//Code for Windows
int App::get_next_file_num()
   {
   static int i = 0;

   if (i == 0)
      {
      WIN32_FIND_DATA ffd;
      HANDLE hFind = INVALID_HANDLE_VALUE;

      char * file_spec = "fracter-???.bmp";
      hFind = FindFirstFile(file_spec, &ffd);
      if (hFind != INVALID_HANDLE_VALUE)
	 {
	 std::string s;
	 std::stringstream ss;
	 do
	    {
	    std::cerr << ffd.cFileName << std::endl;
	    ss << ffd.cFileName;
	    if (ss.str().size() == 18)
	       {
	       s = ss.str().substr(0, 11);
	       if ( StrCmpI(s.c_str(), "fracter-") == 0 )
		  {
		  s = ss.str().substr(14, 4);
		  if (StrCmpI(s.c_str(), ".bmp") == 0)
		     {
		     s = ss.str().substr(11, 3);
		     int j = atoi(s.c_str());
		     if (j > i) i = j;
		     }
		  }
	       }
	    }
	 while (FindNextFile(hFind, &ffd) != 0);
	 }
      }

   i++;
   return i;
   }



#else
////////////////////////////////////////////////////////////////////////////////
// Code for Linux

int App::scandir_filter(const struct dirent * d)
   {
   if (memcmp(d->d_name, "fracter-", 8) != 0) return 0;
   if (memcmp(d->d_name+11, ".bmp", 4) != 0) return 0;
   if (isdigit(d->d_name[8]) 
       && isdigit(d->d_name[9]) 
       && isdigit(d->d_name[10])) 
      return 1;
   return 0;
   }


int App::get_next_file_num()
   {
   static int i = 0;

   if (i == 0)
      {
      struct dirent ** file_list;
      int num_files = scandir(".", &file_list, scandir_filter, alphasort);
      if (num_files != 0)
	 sscanf(file_list[num_files-1]->d_name, "fracter-%03d.bmp", &i);
      }
   i++;

   return i;
   }

#endif

